unit uJoint;

interface

uses
  uBody, uVect, math, snail, windows;
type
  cpJointType = (PIVOT_JOINT, PIN_JOINT, SLIDE_JOINT,DAMPED_SPRING);

  PPivotJoint = ^TPivotJoint;
  TPivotJoint = record
    anchr1, anchr2: cpVect;
    r1, r2: cpVect;
    k1, k2: cpVect;
    jAcc, jBias: cpVect;
    bias: cpVect;
  end;

  PDampedSpring = ^TDampedSpring;
  TDampedSpring = record
	  anchr1, anchr2: cpVect;
	  restLength, stiffness, damping, dt, target_vrn, nMass: single;
    r1, r2, n: cpVect;
  end;

  PPinJoint = ^TPinJoint;
  TPinJoint = record
    anchr1, anchr2: cpVect;
    dist: single;
    r1, r2: cpVect;
    n: cpVect;
    nMass: single;
    jnAcc, jBias: single;
    bias: single;
  end;

  PSlideJoint = ^TSlideJoint;
  TSlideJoint = record
    anchr1, anchr2: cpVect;
    min, max: single;
    r1, r2: cpVect;
    n: cpVect;
    nMass: single;
    jnAcc, jBias: single;
    bias: single;
  end;

  PGrooveJoint = ^TGrooveJoint;
  TGrooveJoint = record
    grv_n, grv_a, grv_b: cpVect;
    anchr2: cpVect;
    grv_tn: cpVect;
    clamp: single;
    r1, r2: cpVect;
    k1, k2: cpVect;
    jAcc, jBias: cpVect;
    bias: cpVect;
  end;

  PJoint = ^TJoint;
  TJoint = record
    JointType: cpJointType;
    body1, body2: PBody;
    DfJoint: pointer;
    Next, Prev: PJoint;
  end;

const
  cp_joint_bias_coef = 0.1;

var
  cpJoint: PJoint;
  JointsCount: integer;

  procedure cpJointDelete(jnt: PJoint);
  function cpPivotJointCreate(body1, body2: PBody; pivot: cpVect): PJoint;
  function cpPinJointCreate(body1, body2: PBody; anchr1, anchr2: cpVect): PJoint;
  function cpSlideJointCreate(body1, body2 : PBody; anchr1, anchr2 : cpVect; min, max : single): PJoint;
  function cpDampedSpringCreate(body1, body2: PBody; anchr1, anchr2: cpVect; restLength, stiffness, damping: single): PJoint;
  procedure JointPreStep(jnt: PJoint; dt_inv: single);
  procedure JointApplyImpulse(jnt: PJoint);
  
implementation

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpJointCreate(JointType: cpJointType; body1, body2: PBody): PJoint;
var
  j: PJoint;
begin
  New(j);
  j.Next:= nil;
  j.Prev:= nil;

  j.JointType:= JointType;
  j.body1:= body1;
  j.body2:= body2;

  j.Next:= cpJoint;
  if cpJoint <> nil then cpJoint.Prev:= j;
  cpJoint:= j;
  Result:= cpJoint;

  inc(JointsCount);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure cpJointDelete(jnt: PJoint);
var
  DelJoint: PJoint;
begin
  DelJoint:= jnt;
  if jnt.Prev <> nil then jnt.Prev.Next := jnt.Next
    else cpJoint:= jnt.Next;
  if jnt.Next <> nil then jnt.Next.Prev := jnt.Prev;

  jnt.body1:= nil;
  jnt.body2:= nil;
  Dispose(jnt.DfJoint);
  Dispose(DelJoint);
  dec(JointsCount);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure PivotJointPreStep(jnt: PJoint; dt_inv: single);
var
  joint: PPivotJoint;
  b1, b2: PBody;
  k11, k12, k21, k22, m_sum, r1xsq, r1ysq, r1nxy, r2xsq, r2ysq, r2nxy, det_inv: single;
  delta: cpVect;
begin
  Joint:= jnt.DfJoint;
	b1:= jnt.body1;
	b2:= jnt.body2;

	joint.r1:= cpvrotate(joint.anchr1, b1.rot);
	joint.r2:= cpvrotate(joint.anchr2, b2.rot);

	m_sum:= b1.m_inv + b2.m_inv;
	k11:= m_sum; k12:= 0;
	k21:= 0;  k22:= m_sum;

	r1xsq:=  joint.r1.x * joint.r1.x * b1.i_inv;
	r1ysq:=  joint.r1.y * joint.r1.y * b1.i_inv;
	r1nxy:= -joint.r1.x * joint.r1.y * b1.i_inv;
	k11:= k11 + r1ysq; k12:= k12 + r1nxy;
	k21:= k21 + r1nxy; k22:= k22 + r1xsq;

	r2xsq:=  joint.r2.x * joint.r2.x * b2.i_inv;
	r2ysq:=  joint.r2.y * joint.r2.y * b2.i_inv;
	r2nxy:= -joint.r2.x * joint.r2.y * b2.i_inv;
	k11:= k11 + r2ysq; k12:= k12 + r2nxy;
	k21:= k21 + r2nxy; k22:= k22 + r2xsq;

	det_inv:= 1.0 / (k11 * k22 - k12 * k21);
	joint.k1:= cpv( k22 * det_inv, -k12 * det_inv);
	joint.k2:= cpv(-k21 * det_inv,  k11 * det_inv);

	delta:= cpvsub(cpvadd(b2.p, joint.r2), cpvadd(b1.p, joint.r1));
	joint.bias:= cpvmult(delta, -cp_joint_bias_coef * dt_inv);
	joint.jBias:= cpv(0, 0);
	
	cpBodyApplyImpulse(b1, cpvneg(joint.jAcc), joint.r1);
	cpBodyApplyImpulse(b2, joint.jAcc, joint.r2);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure PivotJointApplyImpulse(jnt: PJoint);
var
  joint: PPivotJoint;
  b1, b2: PBody;
  r1, r2, k1, k2, vb1, vb2, vbr, jb, v1, v2, vr, j: cpVect;
begin
  joint:= jnt.DfJoint;
	b1:= jnt.body1;
	b2:= jnt.body2;

	r1:= joint.r1;
	r2:= joint.r2;
	k1:= joint.k1;
	k2:= joint.k2;

	vb1:= cpvadd(b1.v_bias, cpvmult(cpvperp(r1), b1.w_bias));
	vb2:= cpvadd(b2.v_bias, cpvmult(cpvperp(r2), b2.w_bias));
	vbr:= cpvsub(joint.bias, cpvsub(vb2, vb1));
	
	jb:= cpv(cpvdot(vbr, k1), cpvdot(vbr, k2));
	joint.jBias:= cpvadd(joint.jBias, jb);
	
	cpBodyApplyBiasImpulse(b1, cpvneg(jb), r1);
	cpBodyApplyBiasImpulse(b2, jb, r2);
	
	v1:= cpvadd(b1.v, cpvmult(cpvperp(r1), b1.w));
	v2:= cpvadd(b2.v, cpvmult(cpvperp(r2), b2.w));
	vr:= cpvsub(v2, v1);
	
	j:= cpv(-cpvdot(vr, k1), -cpvdot(vr, k2));
	joint.jAcc:= cpvadd(joint.jAcc, j);

	cpBodyApplyImpulse(b1, cpvneg(j), r1);
	cpBodyApplyImpulse(b2, j, r2);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpPivotJointCreate(body1, body2: PBody; pivot: cpVect): PJoint;
var
  j: PJoint;
  sj: PPivotJoint;
begin
  New(sj);

  j:= cpJointCreate(PIVOT_JOINT, body1, body2);
  j.DfJoint:= sj;

	sj.anchr1:= cpvunrotate(cpvsub(pivot, body1.p), body1.rot);
	sj.anchr2:= cpvunrotate(cpvsub(pivot, body2.p), body2.rot);
	sj.jAcc:= cpv(0, 0);

  Result:= j;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function k_scalar(a, b: PBody; r1, r2, n: cpVect): single;
var
  mass_sum, r1cn, r2cn: single;
begin
	mass_sum:= a.m_inv + b.m_inv;
	r1cn:= cpvcross(r1, n);
	r2cn:= cpvcross(r2, n);

	result:= mass_sum + a.i_inv * r1cn * r1cn + b.i_inv * r2cn * r2cn;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function relative_velocity(a, b: PBody; r1, r2: cpVect): cpVect;
var
  v1_sum, v2_sum: cpVect;
begin
	v1_sum:= cpvadd(a.v, cpvmult(cpvperp(r1), a.w));
	v2_sum:= cpvadd(b.v, cpvmult(cpvperp(r2), b.w));

	result:= cpvsub(v2_sum, v1_sum);
end;

function relative_velocity_bias(a, b: PBody; r1, r2: cpVect): cpVect;
var
  v1_sum, v2_sum:cpVect;
begin
	v1_sum := cpvadd(a.v_bias, cpvmult(cpvperp(r1), a.w_bias));
	v2_sum := cpvadd(a.v_bias, cpvmult(cpvperp(r2), b.w_bias));

	Result := cpvsub(v2_sum, v1_sum);
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure DampedSpringPreStep(jnt: PJoint; dt_inv: single);
var
  joint: PDampedSpring;
  b1, b2: PBody;
  delta, j: cpVect;
  dist, f_spring: single;
begin
  joint:= jnt.DfJoint;
	b1:= jnt.body1;
	b2:= jnt.body2;
	
	joint.r1:= cpvrotate(joint.anchr1, b1.rot);
	joint.r2:= cpvrotate(joint.anchr2, b2.rot);

	delta:= cpvsub(cpvadd(b2.p, joint.r2), cpvadd(b1.p, joint.r1));
	dist:= cpvlength(delta);


  if dist = 0 then
    joint.n:= cpvmult(delta, 1.0 / INFINITY)
  else
    joint.n:= cpvmult(delta, 1.0 / dist);
	
	joint.nMass:= 1.0 / k_scalar(b1, b2, joint.r1, joint.r2, joint.n);

	joint.dt:= 1.0 / dt_inv;
	joint.target_vrn:= 0;

	f_spring:= (joint.restLength - dist) * joint.stiffness;

  j:= cpvmult(joint.n, f_spring * (1.0 / dt_inv));
	cpBodyApplyImpulse(b1, cpvneg(j), joint.r1);
	cpBodyApplyImpulse(b2, j, joint.r2);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure DampedSpringApplyImpulse(jnt: PJoint);
var
  joint: PDampedSpring;
  b1, b2: PBody;
  n, r1, r2, j: cpVect;
  vrn, v_damp: single;
begin
  joint:= jnt.DfJoint;
	b1:= jnt.body1;
	b2:= jnt.body2;
	
	n:= joint.n;
	r1:= joint.r1;
  r2:= joint.r2;

	vrn:= cpvdot(relative_velocity(b1, b2, r1, r2), n) - joint.target_vrn;

	v_damp:= -vrn * (1 - exp(-joint.damping * joint.dt / joint.nMass));
	joint.target_vrn:= vrn + v_damp;

  j:= cpvmult(joint.n, v_damp * joint.nMass);
  cpBodyApplyImpulse(b1, cpvneg(j), joint.r1);
	cpBodyApplyImpulse(b2, j, joint.r2);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpDampedSpringCreate(body1, body2: PBody; anchr1, anchr2: cpVect; restLength, stiffness, damping: single): PJoint;
var
  j: PJoint;
  sj: PDampedSpring;
begin
  New(sj);

  j:= cpJointCreate(DAMPED_SPRING, body1, body2);
  j.DfJoint:= sj;

	sj.anchr1:= anchr1;
	sj.anchr2:= anchr2;

	sj.restLength:= restLength;
	sj.stiffness:= stiffness;
	sj.damping:= damping;

  Result:= j;
end;

procedure PinJointPreStep(jnt: PJoint; dt_inv: single);
var
  joint: PPinJoint;
  b1, b2: PBody;
  delta, j: cpVect;
  dist: single;
begin
  joint := jnt.DfJoint;
  b1 := jnt.body1;
  b2 := jnt.body2;

  joint.r1 := cpvrotate(joint.anchr1, b1.rot);
  joint.r2 := cpvrotate(joint.anchr2, b2.rot);

  delta := cpvsub(cpvadd(b2.p, joint.r2), cpvadd(b1.p, joint.r1));
  dist := cpvlength(delta);

  if dist = 0 then
    joint.n := cpvmult(delta, 1.0 / Infinity)
  else
    joint.n := cpvmult(delta, 1.0 / dist);


  joint.nMass := 1.0/k_scalar(b1, b2, joint.r1, joint.r2, joint.n);
  joint.bias := -cp_joint_bias_coef*dt_inv*(dist - joint.dist);
  joint.jBias := 0.0;

  j := cpvmult(joint.n, joint.jnAcc);

  cpBodyApplyImpulse(b1, cpvneg(j), joint.r1);
	cpBodyApplyImpulse(b2, j, joint.r2);
end;

procedure PinJointApplyImpulse(jnt: PJoint);
var
  joint: PPinJoint;
  b1, b2: PBody;
  n, r1, r2, jb, j: cpVect;
  vbn, jbn, vrn, jn: single;
begin
  joint := jnt.DfJoint;
  b1 := jnt.body1;
  b2 := jnt.body2;

  n := joint.n;
  r1 := joint.r1;
  r2 := joint.r2;

  vbn := cpvdot(relative_velocity_bias(b1, b2, r1, r2), n);

  jbn := (joint.bias - vbn) * joint.nMass;
  joint.jBias := joint.jBias + jbn;

  jb := cpvmult(n, jbn);
  cpBodyApplyBiasImpulse(b1, cpvneg(jb), r1);
	cpBodyApplyBiasImpulse(b2, jb, r2);

  vrn := cpvdot(relative_velocity(b1, b2, r1 ,r2), n);

  jn := -vrn * joint.nMass;
  joint.jnAcc := jn;

  j := cpvmult(n, jn);
  cpBodyApplyImpulse(b1, cpvneg(j), r1);
	cpBodyApplyImpulse(b2, j, r2);
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpPinJointCreate(body1, body2: PBody; anchr1, anchr2: cpVect): PJoint;
var
  j: PJoint;
  sj: PPinJoint;
  p1, p2: cpVect;
begin
  New(sj);

  j := cpJointCreate(PIN_JOINT, body1, body2);
  j.DfJoint := sj;

  sj.anchr1 := anchr1;
  sj.anchr2 := anchr2;

  p1 := cpvadd(body1.p, cpvrotate(anchr1, body1.rot));
  p2 := cpvadd(body2.p, cpvrotate(anchr2, body2.rot));
  sj.dist := cpvlength(cpvsub(p2, p1));
  sj.jnAcc := 0.0;

  Result := j;
end;

procedure SlideJointPreStep(jnt: PJoint; dt_inv: single);
var
  b1, b2: PBody;
  joint: PSlideJoint;
  delta, j: cpVect;
  dist, pdist: single;
begin
  joint := jnt.DfJoint;
  b1 := jnt.body1;
  b2 := jnt.body2;

  joint.r1 := cpvrotate(joint.anchr1, b1.rot);
  joint.r2 := cpvrotate(joint.anchr2, b2.rot);

  delta := cpvsub(cpvadd(b2.p, joint.r2), cpvadd(b1.p, joint.r1));
  dist := cpvlength(delta);
  pdist := 0.0;

  if dist > joint.max then
    pdist := dist - joint.max
  else
    if dist < joint.min then
    begin
      pdist := joint.min - dist;
      dist := -dist;
    end;

  if dist = 0 then
    joint.n := cpvmult(delta, 1.0 / Infinity)
  else
    joint.n := cpvmult(delta, 1.0 / dist);

  joint.nMass := 1.0/k_scalar(b1, b2, joint.r1, joint.r2, joint.n);
  joint.bias := -cp_joint_bias_coef * dt_inv * pdist;
  joint.jBias := 0.0;

  if joint.bias = 0 then
    joint.jnAcc := 0.0;

  j := cpvmult(joint.n, joint.jnAcc);

  cpBodyApplyImpulse(b1, cpvneg(j), joint.r1);
	cpBodyApplyImpulse(b2, j, joint.r2);
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure SlideJointApplyImpulse(jnt:PJoint);
var
  joint: PSlideJoint;
  b1, b2: PBody;
  n, r1, r2, jb, j: cpVect;
  vbn, jbn, jbnOld, vrn, jn, jnOld: single;
begin
  joint := jnt.DfJoint;
  if joint.bias = 0 then exit;

  b1 := jnt.body1;
  b2 := jnt.body2;

  n := joint.n;
  r1 := joint.r1;
  r2 := joint.r2;

  vbn := cpvdot(relative_velocity_bias(b1, b2, r1, r2), n);

  jbn := (joint.bias - vbn) * joint.nMass;
  jbnOld := joint.jBias;
  joint.jBias := min(jbnOld + jbn, 0.0);
  jbn := joint.jBias - jbnOld;

  jb := cpvmult(n, jbn);
  cpBodyApplyBiasImpulse(b1, cpvneg(jb), r1);
	cpBodyApplyBiasImpulse(b2, jb, r2);

  vrn := cpvdot(relative_velocity(b1, b2, r1, r2), n);

  jn := -vrn*joint.nMass;
  jnOld := joint.jnAcc;
  joint.jnAcc := min(jnOld + jn, 0.0);
  jn := joint.jnAcc - jnOld;

  j := cpvmult(n, jn);

  cpBodyApplyImpulse(b1, cpvneg(j), joint.r1);
	cpBodyApplyImpulse(b2, j, joint.r2);
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
function cpSlideJointCreate(body1, body2 : PBody; anchr1, anchr2 : cpVect; min, max : single): PJoint;
var
  j: PJoint;
  sj: PSlideJoint;
begin
  New(sj);

  j := cpJointCreate(SLIDE_JOINT, body1, body2);
  j.DfJoint := sj;

  sj.anchr1 := anchr1;
  sj.anchr2 := anchr2;

  sj.min := min;
  sj.max := max;
  sj.jnAcc := 0.0;

  Result := j;
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
{procedure GrooveJointPreStep(jnt : PJoint; dt_inv : single);
var
  joint : PGrooveJoint;
  b1, b2 : PBody;
  ta, tb, n : cpVect;
  d, td, k11, k12, k21, k22, m_sum, r1xsq, r1ysq, r1nxy, det_inv : single;
begin
  joint := jnt.DfJoint;
  b1 := jnt.body1;
  b2 := jnt.body2;

  ta := cpvadd(b1.p, cpvrotate(joint.grv_a, b1.rot));
  tb := cpvadd(b1.p, cpvrotate(joint.grv_b, b1.rot));

  n := cpvrotate(joint.grv_n, b1.rot);
  d := cpvdot(ta, n);

  joint.grv_tn := n;
  joint.r2 := cpvrotate(joint.anchr2, b2.rot);

  td := cpvcross(cpvadd(b2.p, joint.r2), n);
  if td <= cpvcross(ta, n) then
  begin
    joint.clamp := 1.0;
    joint.r1 := cpvsub(ta, b1.p);
  end else
    if td >= cpvcross(tb, n) then
    begin
      joint.clamp := -1.0;
      joint.r1 := cpvsub(tb, b1.p);
    end else
    begin
      joint.clamp := 0.0;
      joint.r1 := cpvsub(cpvadd(cpvmult(cpvperp(n), -td), cpvmult(n, d)), b1.p);
    end;

  m_sum := b1.m_inv + b2.m_inv;
  k11 := m_sum;
  k12 := 0.0;
	k21 := 0.0;
  k22 := m_sum;

  r1xsq :=  joint.r1.x * joint.r1.x * b1.i_inv;
	r1ysq :=  joint.r1.y * joint.r1.y * b1.i_inv;
	r1nxy := -joint.r1.x * joint.r1.y * b1.i_inv;
	k11 := k11 + r1ysq;
  k12 := k12 + r1nxy;
	k21 := k21 + r1nxy;
  k22 := k22 + r1xsq;

  r1xsq :=  joint.r2.x * joint.r2.x * b2.i_inv;
	r1ysq :=  joint.r2.y * joint.r2.y * b2.i_inv;
	r1nxy := -joint.r2.x * joint.r2.y * b2.i_inv;
	k11 := k11 + r1ysq;
  k12 := k12 + r1nxy;
	k21 := k21 + r1nxy;
  k22 := k22 + r1xsq;

	det_inv := 1.0/(k11*k22 - k12*k21);
	joint.k1 := cpv( k22*det_inv, -k12*det_inv);
	joint.k2 := cpv(-k21*det_inv,  k11*det_inv);

	ta := cpvsub(cpvadd(b2.p, joint.r2), cpvadd(b1.p, joint.r1));
	joint.bias := cpvmult(ta, -cp_joint_bias_coef * dt_inv);
	joint.jBias := cpvzero;

  cpBodyApplyImpulse(b1, cpvneg(joint.jAcc), joint.r1);
	cpBodyApplyImpulse(b2, joint.jAcc, joint.r2);
end;
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
}
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure JointPreStep(jnt: PJoint; dt_inv: single);
begin
  case jnt.JointType of
    PIVOT_JOINT : pivotJointPreStep(jnt, dt_inv);
    PIN_JOINT : PinJointPreStep(jnt, dt_inv);
    SLIDE_JOINT : SlideJointPreStep(jnt, dt_inv);
    DAMPED_SPRING : DampedSpringPreStep(jnt, dt_inv);
  end;
end;

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
procedure JointApplyImpulse(jnt: PJoint);
begin
  case jnt.JointType of
    PIVOT_JOINT : pivotJointApplyImpulse(jnt);
    PIN_JOINT : PinJointApplyImpulse(jnt);
    SLIDE_JOINT : SlideJointApplyImpulse(jnt);
    DAMPED_SPRING : DampedSpringApplyImpulse(jnt);
  end;
end;

end.
